from langchain import hub
from langchain.agents import AgentExecutor, create_openai_tools_agent, load_tools
import gradio as gr
from gradio_agentchatbot import AgentChatbot, Message

from langchain_openai import ChatOpenAI

from dotenv import load_dotenv

load_dotenv()

model = ChatOpenAI(temperature=0, streaming=True)

tools = load_tools(["serpapi"])

prompt = hub.pull("hwchase17/openai-tools-agent")
agent = create_openai_tools_agent(
    model.with_config({"tags": ["agent_llm"]}), tools, prompt
)
langchain_agent = AgentExecutor(agent=agent, tools=tools).with_config(
    {"run_name": "Agent"}
)


async def interact_with_langchain_agent(prompt, messages):
    messages.append(Message(role="user", content=prompt))
    yield messages
    async for chunk in langchain_agent.astream({"input": prompt}):
        if "steps" in chunk:
            for step in chunk["steps"]:
                messages.append(
                    dict(
                        role="assistant",
                        content=step.action.log,
                        metadata={"tool_name": step.action.tool},
                    )
                )
                yield messages
        if "output" in chunk:
            messages.append(Message(role="assistant", content=chunk["output"]))
            yield messages


with gr.Blocks() as demo:
    gr.Markdown("# Chat with a LangChain Agent 🦜⛓️ and see its thoughts 💭")
    chatbot_2 = AgentChatbot(
        label="Agent",
        avatar_images=(
            None,
            "https://em-content.zobj.net/source/twitter/141/parrot_1f99c.png",
        ),
    )
    input_2 = gr.Textbox(lines=1, label="Chat Message")
    input_2.submit(interact_with_langchain_agent, [input_2, chatbot_2], [chatbot_2])

if __name__ == "__main__":
    demo.launch()



from huggingface_hub import InferenceClient
from gradio_agentchatbot import Message, AgentChatbot
import gradio as gr

"""
For more information on `huggingface_hub` Inference API support, please check the docs: https://huggingface.co/docs/huggingface_hub/v0.22.2/en/guides/inference
"""
client = InferenceClient("HuggingFaceH4/zephyr-7b-beta")


def respond(
    prompt: str,
    history,
):
    if not history:
        history = [{"role": "system", "content": "You are a friendly chatbot"}]
    history.append({"role": "user", "content": prompt})

    yield history

    response = {"role": "assistant", "content": ""}
    for message in client.chat_completion(
        history,
        stream=True,
    ):
        response["content"] += message.choices[0].delta.content or ""

        yield history + [response]


with gr.Blocks() as demo:
    gr.Markdown("# Chat with Hugging Face Zephyr 7b 🤗")
    chatbot = AgentChatbot(
        label="Agent",
        avatar_images=(
            None,
            "https://em-content.zobj.net/source/twitter/376/hugging-face_1f917.png",
        ),
    )
    prompt = gr.Textbox(lines=1, label="Chat Message")
    prompt.submit(respond, [prompt, chatbot], [chatbot])


if __name__ == "__main__":
    demo.launch()



from typing_extensions import NotRequired

from transformers.agents import ReactAgent, agent_types
from pydantic import Field
from gradio.data_classes import GradioModel, FileData, GradioRootModel
from typing import Literal, List, Generator, Optional, Protocol, TypedDict, Union, Any


class MetadataDict(TypedDict):
    tool_name: Union[str, None]
    error: bool


class FileDataDict(TypedDict):
    path: str  # server filepath
    url: Optional[str]  # normalised server url
    size: Optional[int]  # size in bytes
    orig_name: Optional[str]  # original filename
    mime_type: Optional[str]
    is_stream: bool
    meta: dict[Literal["_type"], Literal["gradio.FileData"]]


class MessageDict(TypedDict):
    content: str | FileDataDict
    role: Literal["user", "assistant", "system"]
    metadata: NotRequired[MetadataDict]


class ApiReturnObj(Protocol):
    @property
    def content(self) -> str: ...

    @property
    def role(self) -> Literal["user", "assistant", "system", "tool"]: ...


class Metadata(GradioModel):
    tool_name: Optional[str] = None
    error: bool = False


class FileMessage(GradioModel):
    file: FileData
    alt_text: Optional[str] = None


class Message(GradioModel):
    role: Literal["user", "assistant", "system", "tool"]
    metadata: Metadata = Field(default_factory=Metadata)
    content: str | FileMessage

    @staticmethod
    def matches_protocol(obj: Any):
        return hasattr(obj, "role") and hasattr(obj, "content")

    @classmethod
    def from_api(cls, obj: ApiReturnObj) -> "Message":
        return cls(role=obj.role, content=obj.content or "")

    def append(self, obj: ApiReturnObj):
        if isinstance(self.content, FileMessage):
            raise ValueError("Cannot append to a message with a file.")
        self.content += obj.content or ""
        return self


class ChatbotData(GradioRootModel):
    root: List[Message]


def convert_to_message_stream(message: dict) -> Generator[Message, None, None]:
    if message.get("rationale"):
        yield Message(role="assistant", content=message["rationale"])
    if message.get("tool_call"):
        used_code = message["tool_call"]["tool_name"] == "code interpreter"
        content = message["tool_call"]["tool_arguments"]
        if used_code:
            content = f"```py\n{content}\n```"
        yield Message(
            role="assistant",
            metadata=Metadata(tool_name=message["tool_call"]["tool_name"]),
            content=content,
        )
    if message.get("observation"):
        yield Message(role="assistant", content=message["observation"])
    if message.get("error"):
        yield Message(
            role="assistant",
            content=str(message["error"]),
            metadata=Metadata(error=True),
        )


def pull_messages(new_messages: List[dict]):
    for message in new_messages:
        if not len(message):
            continue
        if message.get("rationale"):
            yield Message(role="assistant", content=message["rationale"])
        if message.get("tool_call"):
            used_code = message["tool_call"]["tool_name"] == "code interpreter"
            content = message["tool_call"]["tool_arguments"]
            if used_code:
                content = f"```py\n{content}\n```"
            yield Message(
                role="assistant",
                metadata=Metadata(tool_name=message["tool_call"]["tool_name"]),
                content=content,
            )
        if message.get("observation"):
            yield Message(role="assistant", content=message["observation"])
        if message.get("error"):
            yield Message(
                role="assistant",
                content=str(message["error"]),
                metadata=Metadata(error=True),
            )


def stream_from_transformers_agent(
    agent: ReactAgent, prompt: str
) -> Generator[Message, None, None]:
    """Runs an agent with the given prompt and streams the messages from the agent as ChatMessages."""

    for message in agent.run(prompt, stream=True):  # type: ignore
        if isinstance(message, dict):
            for gradio_message in convert_to_message_stream(message):
                yield gradio_message
        elif isinstance(message, agent_types.AgentText):
            yield Message(role="assistant", content=message.to_string())
        elif isinstance(message, agent_types.AgentImage):
            yield Message(
                role="assistant",
                content=FileMessage(
                    file=FileData(path=message.to_string(), mime_type="image/png")
                ),  # type: ignore
            )
        elif isinstance(message, agent_types.AgentAudio):
            yield Message(
                role="assistant",
                content=FileMessage(
                    file=FileData(path=message.to_string(), mime_type="audio/wav")
                ),  # type: ignore
            )
        elif isinstance(message, str):
            yield Message(role="assistant", content=message)



"""
This file defines a useful high-level abstraction to build Gradio chatbots: ChatInterface.
"""

from __future__ import annotations

import inspect
from typing import AsyncGenerator, Generator, Callable, Literal, Union, cast, Protocol

import anyio
from gradio_client.documentation import document

from gradio.blocks import Blocks
from gradio.components import (
    Button,
    Chatbot,
    Component,
    Markdown,
    MultimodalTextbox,
    State,
    Textbox,
    get_component_instance,
)
from gradio.components.multimodal_textbox import MultimodalData
from gradio.events import Dependency, on
from gradio.helpers import create_examples as Examples  # noqa: N812
from gradio.helpers import special_args
from gradio.layouts import Accordion, Group, Row
from gradio.routes import Request
from gradio.themes import ThemeClass as Theme
from gradio.utils import SyncToAsyncIterator, async_iteration, async_lambda
from .utils import MessageDict, Message, ApiReturnObj, FileDataDict
from .agentchatbot import AgentChatbot
from gradio.exceptions import Error

History = list[MessageDict | Message | ApiReturnObj]


class ChatInterfaceFn(Protocol):
    def __call__(
        self, message: str, history: list[MessageDict], *args, **kwargs
    ) -> MessageDict | ApiReturnObj | Message: ...


class ChatInterfaceGeneratorFn(Protocol):
    def __call__(
        self, message: str, history: list[MessageDict], *args, **kwargs
    ) -> Generator[MessageDict | ApiReturnObj | Message, None, None]: ...


class ChatInterfaceAsyncGeneratorFn(Protocol):
    async def __call__(
        self, message: str, history: list[MessageDict], *args, **kwargs
    ) -> AsyncGenerator[MessageDict | ApiReturnObj | Message, None]: ...


class ChatInterfaceAsyncFn(Protocol):
    async def __call__(
        self, message: str, history: list[MessageDict], *args, **kwargs
    ) -> MessageDict | ApiReturnObj | Message: ...


@document()
class ChatInterface(Blocks):
    """
    ChatInterface is Gradio's high-level abstraction for creating chatbot UIs, and allows you to create
    a web-based demo around a chatbot model in a few lines of code. Only one parameter is required: fn, which
    takes a function that governs the response of the chatbot based on the user input and chat history. Additional
    parameters can be used to control the appearance and behavior of the demo.

    Example:
        import gradio as gr

        def echo(message, history):
            return message

        demo = gr.ChatInterface(fn=echo, examples=["hello", "hola", "merhaba"], title="Echo Bot")
        demo.launch()
    Demos: chatinterface_multimodal, chatinterface_random_response, chatinterface_streaming_echo
    Guides: creating-a-chatbot-fast, sharing-your-app
    """

    def __init__(
        self,
        fn: ChatInterfaceFn
        | ChatInterfaceGeneratorFn
        | ChatInterfaceAsyncFn
        | ChatInterfaceAsyncGeneratorFn,
        *,
        multimodal: bool = False,
        chatbot: Chatbot | None = None,
        textbox: Textbox | MultimodalTextbox | None = None,
        additional_inputs: str | Component | list[str | Component] | None = None,
        additional_inputs_accordion_name: str | None = None,
        additional_inputs_accordion: str | Accordion | None = None,
        examples: list[str] | list[dict[str, str | list]] | list[list] | None = None,
        cache_examples: bool | Literal["lazy"] | None = None,
        examples_per_page: int = 10,
        title: str | None = None,
        description: str | None = None,
        theme: Theme | str | None = None,
        css: str | None = None,
        js: str | None = None,
        head: str | None = None,
        analytics_enabled: bool | None = None,
        submit_btn: str | None | Button = "Submit",
        stop_btn: str | None | Button = "Stop",
        retry_btn: str | None | Button = "🔄  Retry",
        undo_btn: str | None | Button = "↩️ Undo",
        clear_btn: str | None | Button = "🗑️  Clear",
        autofocus: bool = True,
        concurrency_limit: int | None | Literal["default"] = "default",
        fill_height: bool = True,
        delete_cache: tuple[int, int] | None = None,
    ):
        """
        Parameters:
            fn: The function to wrap the chat interface around. Should accept two parameters: a string input message and list of two-element lists of the form [[user_message, bot_message], ...] representing the chat history, and return a string response. See the Chatbot documentation for more information on the chat history format.
            multimodal: If True, the chat interface will use a gr.MultimodalTextbox component for the input, which allows for the uploading of multimedia files. If False, the chat interface will use a gr.Textbox component for the input.
            chatbot: An instance of the gr.Chatbot component to use for the chat interface, if you would like to customize the chatbot properties. If not provided, a default gr.Chatbot component will be created.
            textbox: An instance of the gr.Textbox or gr.MultimodalTextbox component to use for the chat interface, if you would like to customize the textbox properties. If not provided, a default gr.Textbox or gr.MultimodalTextbox component will be created.
            additional_inputs: An instance or list of instances of gradio components (or their string shortcuts) to use as additional inputs to the chatbot. If components are not already rendered in a surrounding Blocks, then the components will be displayed under the chatbot, in an accordion.
            additional_inputs_accordion_name: Deprecated. Will be removed in a future version of Gradio. Use the `additional_inputs_accordion` parameter instead.
            additional_inputs_accordion: If a string is provided, this is the label of the `gr.Accordion` to use to contain additional inputs. A `gr.Accordion` object can be provided as well to configure other properties of the container holding the additional inputs. Defaults to a `gr.Accordion(label="Additional Inputs", open=False)`. This parameter is only used if `additional_inputs` is provided.
            examples: Sample inputs for the function; if provided, appear below the chatbot and can be clicked to populate the chatbot input. Should be a list of strings if `multimodal` is False, and a list of dictionaries (with keys `text` and `files`) if `multimodal` is True.
            cache_examples: If True, caches examples in the server for fast runtime in examples. The default option in HuggingFace Spaces is True. The default option elsewhere is False.
            examples_per_page: If examples are provided, how many to display per page.
            title: a title for the interface; if provided, appears above chatbot in large font. Also used as the tab title when opened in a browser window.
            description: a description for the interface; if provided, appears above the chatbot and beneath the title in regular font. Accepts Markdown and HTML content.
            theme: Theme to use, loaded from gradio.themes.
            css: Custom css as a string or path to a css file. This css will be included in the demo webpage.
            js: Custom js as a string or path to a js file. The custom js should be in the form of a single js function. This function will automatically be executed when the page loads. For more flexibility, use the head parameter to insert js inside <script> tags.
            head: Custom html to insert into the head of the demo webpage. This can be used to add custom meta tags, multiple scripts, stylesheets, etc. to the page.
            analytics_enabled: Whether to allow basic telemetry. If None, will use GRADIO_ANALYTICS_ENABLED environment variable if defined, or default to True.
            submit_btn: Text to display on the submit button. If None, no button will be displayed. If a Button object, that button will be used.
            stop_btn: Text to display on the stop button, which replaces the submit_btn when the submit_btn or retry_btn is clicked and response is streaming. Clicking on the stop_btn will halt the chatbot response. If set to None, stop button functionality does not appear in the chatbot. If a Button object, that button will be used as the stop button.
            retry_btn: Text to display on the retry button. If None, no button will be displayed. If a Button object, that button will be used.
            undo_btn: Text to display on the delete last button. If None, no button will be displayed. If a Button object, that button will be used.
            clear_btn: Text to display on the clear button. If None, no button will be displayed. If a Button object, that button will be used.
            autofocus: If True, autofocuses to the textbox when the page loads.
            concurrency_limit: If set, this is the maximum number of chatbot submissions that can be running simultaneously. Can be set to None to mean no limit (any number of chatbot submissions can be running simultaneously). Set to "default" to use the default concurrency limit (defined by the `default_concurrency_limit` parameter in `.queue()`, which is 1 by default).
            fill_height: If True, the chat interface will expand to the height of window.
            delete_cache: A tuple corresponding [frequency, age] both expressed in number of seconds. Every `frequency` seconds, the temporary files created by this Blocks instance will be deleted if more than `age` seconds have passed since the file was created. For example, setting this to (86400, 86400) will delete temporary files every day. The cache will be deleted entirely when the server restarts. If None, no cache deletion will occur.
        """
        super().__init__(
            analytics_enabled=analytics_enabled,
            mode="chat_interface",
            css=css,
            title=title or "Gradio",
            theme=theme,
            js=js,
            head=head,
            fill_height=fill_height,
            delete_cache=delete_cache,
        )
        self.multimodal = multimodal
        self.concurrency_limit = concurrency_limit
        self.fn = fn
        self.is_async = inspect.iscoroutinefunction(
            self.fn
        ) or inspect.isasyncgenfunction(self.fn)
        self.is_generator = inspect.isgeneratorfunction(
            self.fn
        ) or inspect.isasyncgenfunction(self.fn)
        self.buttons: list[Button | None] = []

        self.examples = examples
        self.cache_examples = cache_examples

        if additional_inputs:
            if not isinstance(additional_inputs, list):
                additional_inputs = [additional_inputs]
            self.additional_inputs = [
                get_component_instance(i)
                for i in additional_inputs  # type: ignore
            ]
        else:
            self.additional_inputs = []
        if additional_inputs_accordion_name is not None:
            print(
                "The `additional_inputs_accordion_name` parameter is deprecated and will be removed in a future version of Gradio. Use the `additional_inputs_accordion` parameter instead."
            )
            self.additional_inputs_accordion_params = {
                "label": additional_inputs_accordion_name
            }
        if additional_inputs_accordion is None:
            self.additional_inputs_accordion_params = {
                "label": "Additional Inputs",
                "open": False,
            }
        elif isinstance(additional_inputs_accordion, str):
            self.additional_inputs_accordion_params = {
                "label": additional_inputs_accordion
            }
        elif isinstance(additional_inputs_accordion, Accordion):
            self.additional_inputs_accordion_params = (
                additional_inputs_accordion.recover_kwargs(
                    additional_inputs_accordion.get_config()
                )
            )
        else:
            raise ValueError(
                f"The `additional_inputs_accordion` parameter must be a string or gr.Accordion, not {type(additional_inputs_accordion)}"
            )

        with self:
            if title:
                Markdown(
                    f"<h1 style='text-align: center; margin-bottom: 1rem'>{self.title}</h1>"
                )
            if description:
                Markdown(description)

            if chatbot:
                self.chatbot = chatbot.render()
            else:
                self.chatbot = AgentChatbot(
                    label="Chatbot", scale=1, height=200 if fill_height else None
                )

            with Row():
                for btn in [retry_btn, undo_btn, clear_btn]:
                    if btn is not None:
                        if isinstance(btn, Button):
                            btn.render()
                        elif isinstance(btn, str):
                            btn = Button(
                                btn, variant="secondary", size="sm", min_width=60
                            )
                        else:
                            raise ValueError(
                                f"All the _btn parameters must be a gr.Button, string, or None, not {type(btn)}"
                            )
                    self.buttons.append(btn)  # type: ignore

            with Group():
                with Row():
                    if textbox:
                        if self.multimodal:
                            submit_btn = None
                        else:
                            textbox.container = False
                        textbox.show_label = False
                        textbox_ = textbox.render()
                        if not isinstance(textbox_, (Textbox, MultimodalTextbox)):
                            raise TypeError(
                                f"Expected a gr.Textbox or gr.MultimodalTextbox component, but got {type(textbox_)}"
                            )
                        self.textbox = textbox_
                    elif self.multimodal:
                        submit_btn = None
                        self.textbox = MultimodalTextbox(
                            show_label=False,
                            label="Message",
                            placeholder="Type a message...",
                            scale=7,
                            autofocus=autofocus,
                        )
                    else:
                        self.textbox = Textbox(
                            container=False,
                            show_label=False,
                            label="Message",
                            placeholder="Type a message...",
                            scale=7,
                            autofocus=autofocus,
                        )
                    if submit_btn is not None and not multimodal:
                        if isinstance(submit_btn, Button):
                            submit_btn.render()
                        elif isinstance(submit_btn, str):
                            submit_btn = Button(
                                submit_btn,
                                variant="primary",
                                scale=1,
                                min_width=150,
                            )
                        else:
                            raise ValueError(
                                f"The submit_btn parameter must be a gr.Button, string, or None, not {type(submit_btn)}"
                            )
                    if stop_btn is not None:
                        if isinstance(stop_btn, Button):
                            stop_btn.visible = False
                            stop_btn.render()
                        elif isinstance(stop_btn, str):
                            stop_btn = Button(
                                stop_btn,
                                variant="stop",
                                visible=False,
                                scale=1,
                                min_width=150,
                            )
                        else:
                            raise ValueError(
                                f"The stop_btn parameter must be a gr.Button, string, or None, not {type(stop_btn)}"
                            )
                    self.buttons.extend([submit_btn, stop_btn])  # type: ignore

                self.fake_api_btn = Button("Fake API", visible=False)
                self.fake_response_textbox = Textbox(label="Response", visible=False)
                (
                    self.retry_btn,
                    self.undo_btn,
                    self.clear_btn,
                    self.submit_btn,
                    self.stop_btn,
                ) = self.buttons

            if examples:
                if self.is_generator:
                    examples_fn = self._examples_stream_fn
                else:
                    examples_fn = self._examples_fn

                self.examples_handler = Examples(
                    examples=examples,
                    inputs=[self.textbox] + self.additional_inputs,
                    outputs=self.chatbot,
                    fn=examples_fn,
                    cache_examples=self.cache_examples,
                    _defer_caching=True,
                    examples_per_page=examples_per_page,
                )

            any_unrendered_inputs = any(
                not inp.is_rendered for inp in self.additional_inputs
            )
            if self.additional_inputs and any_unrendered_inputs:
                with Accordion(**self.additional_inputs_accordion_params):  # type: ignore
                    for input_component in self.additional_inputs:
                        if not input_component.is_rendered:
                            input_component.render()

            # The example caching must happen after the input components have rendered
            if examples:
                self.examples_handler._start_caching()

            self.saved_input = State()
            self.chatbot_state = (
                State(self.chatbot.value) if self.chatbot.value else State([])
            )

            self._setup_events()
            self._setup_api()

    def _setup_events(self) -> None:
        submit_fn = self._stream_fn if self.is_generator else self._submit_fn
        submit_triggers = (
            [self.textbox.submit, self.submit_btn.click]
            if self.submit_btn
            else [self.textbox.submit]
        )
        submit_event = (
            on(
                submit_triggers,
                self._clear_and_save_textbox,
                [self.textbox],
                [self.textbox, self.saved_input],
                show_api=False,
                queue=False,
                preprocess=False,
            )
            .then(
                self._display_input,
                [self.saved_input, self.chatbot_state],
                [self.chatbot, self.chatbot_state],
                show_api=False,
                queue=False,
            )
            .then(
                submit_fn,
                [self.saved_input, self.chatbot_state] + self.additional_inputs,
                [self.chatbot, self.chatbot_state],
                show_api=False,
                concurrency_limit=cast(
                    Union[int, Literal["default"], None], self.concurrency_limit
                ),
            )
        )
        self._setup_stop_events(submit_triggers, submit_event)

        if self.retry_btn:
            retry_event = (
                self.retry_btn.click(
                    self._delete_prev_fn,
                    [self.saved_input, self.chatbot_state],
                    [self.chatbot, self.saved_input, self.chatbot_state],
                    show_api=False,
                    queue=False,
                )
                .then(
                    self._display_input,
                    [self.saved_input, self.chatbot_state],
                    [self.chatbot, self.chatbot_state],
                    show_api=False,
                    queue=False,
                )
                .then(
                    submit_fn,
                    [self.saved_input, self.chatbot_state] + self.additional_inputs,
                    [self.chatbot, self.chatbot_state],
                    show_api=False,
                    concurrency_limit=cast(
                        Union[int, Literal["default"], None], self.concurrency_limit
                    ),
                )
            )
            self._setup_stop_events([self.retry_btn.click], retry_event)

        if self.undo_btn:
            self.undo_btn.click(
                self._delete_prev_fn,
                [self.saved_input, self.chatbot_state],
                [self.chatbot, self.saved_input, self.chatbot_state],
                show_api=False,
                queue=False,
            ).then(
                async_lambda(lambda x: x),
                [self.saved_input],
                [self.textbox],
                show_api=False,
                queue=False,
            )

        if self.clear_btn:
            self.clear_btn.click(
                async_lambda(lambda: ([], [], None)),
                None,
                [self.chatbot, self.chatbot_state, self.saved_input],
                queue=False,
                show_api=False,
            )

    def _setup_stop_events(
        self, event_triggers: list[Callable], event_to_cancel: Dependency
    ) -> None:
        if self.stop_btn and self.is_generator:
            if self.submit_btn:
                for event_trigger in event_triggers:
                    event_trigger(
                        async_lambda(
                            lambda: (
                                Button(visible=False),
                                Button(visible=True),
                            )
                        ),
                        None,
                        [self.submit_btn, self.stop_btn],
                        show_api=False,
                        queue=False,
                    )
                event_to_cancel.then(
                    async_lambda(lambda: (Button(visible=True), Button(visible=False))),
                    None,
                    [self.submit_btn, self.stop_btn],
                    show_api=False,
                    queue=False,
                )
            else:
                for event_trigger in event_triggers:
                    event_trigger(
                        async_lambda(lambda: Button(visible=True)),
                        None,
                        [self.stop_btn],
                        show_api=False,
                        queue=False,
                    )
                event_to_cancel.then(
                    async_lambda(lambda: Button(visible=False)),
                    None,
                    [self.stop_btn],
                    show_api=False,
                    queue=False,
                )
            self.stop_btn.click(
                None,
                None,
                None,
                cancels=event_to_cancel,
                show_api=False,
            )

    def _setup_api(self) -> None:
        api_fn = self._api_stream_fn if self.is_generator else self._api_submit_fn

        self.fake_api_btn.click(
            api_fn,
            [self.textbox, self.chatbot_state] + self.additional_inputs,
            [self.textbox, self.chatbot_state],
            api_name="chat",
            concurrency_limit=cast(
                Union[int, Literal["default"], None], self.concurrency_limit
            ),
        )

    def _clear_and_save_textbox(
        self, message: str | MultimodalData
    ) -> tuple[str | dict, str | MultimodalData]:
        if self.multimodal:
            return {"text": "", "files": []}, message
        else:
            return "", message

    def _append_multimodal_history(
        self,
        message: MultimodalData,
        response: MessageDict | None,
        history: list[MessageDict],
    ):
        for x in message.files:
            history.append(
                {"role": "user", "content": cast(FileDataDict, x.model_dump())}
            )
        if message.text is None or not isinstance(message.text, str):
            return
        else:
            history.append({"role": "user", "content": message.text})
        if response:
            history.append(response)

    async def _display_input(
        self, message: str | MultimodalData, history: list[MessageDict]
    ) -> tuple[list[MessageDict], list[MessageDict]]:
        if self.multimodal and isinstance(message, MultimodalData):
            self._append_multimodal_history(message, None, history)
        elif isinstance(message, str):
            history.append({"role": "user", "content": message})
        return history, history

    def response_as_dict(
        self,
        response: MessageDict | ApiReturnObj | Message | str,
        current_response: MessageDict | None = None,
    ) -> MessageDict:
        if Message.matches_protocol(response):
            new_response = Message.from_api(cast(ApiReturnObj, response)).model_dump()
        elif isinstance(response, Message):
            new_response = response.model_dump()
        elif isinstance(response, str):
            assert current_response is not None and isinstance(
                current_response["content"], str
            )
            current_response["content"] += response
            new_response = current_response
            print("new_response", new_response)
        else:
            new_response = response
        return cast(MessageDict, new_response)

    async def _submit_fn(
        self,
        message: str | MultimodalData,
        history_with_input: list[MessageDict],
        request: Request,
        *args,
    ) -> tuple[list[MessageDict], list[MessageDict]]:
        if self.multimodal and isinstance(message, MultimodalData):
            remove_input = (
                len(message.files) + 1
                if message.text is not None
                else len(message.files)
            )
            history = history_with_input[:-remove_input]
            message_serialized = message.model_dump()
        else:
            history = history_with_input[:-1]
            message_serialized = message

        inputs, _, _ = special_args(
            self.fn, inputs=[message_serialized, history, *args], request=request
        )

        if self.is_async:
            response = await cast(ChatInterfaceAsyncFn, self.fn)(*inputs)
        else:
            response: (
                ApiReturnObj | Message | MessageDict
            ) = await anyio.to_thread.run_sync(
                cast(ChatInterfaceFn, self.fn), *inputs, limiter=self.limiter
            )
        new_response = self.response_as_dict(response)

        if self.multimodal and isinstance(message, MultimodalData):
            self._append_multimodal_history(
                message, cast(MessageDict, new_response), history
            )
        else:
            history.append(cast(MessageDict, new_response))
        return history, history

    async def _stream_fn(
        self,
        message: str | MultimodalData,
        history_with_input: list[MessageDict],
        request: Request,
        *args,
    ) -> AsyncGenerator:
        if self.multimodal and isinstance(message, MultimodalData):
            remove_input = (
                len(message.files) + 1
                if message.text is not None
                else len(message.files)
            )
            history = history_with_input[:-remove_input]
        else:
            history = history_with_input[:-1]
        inputs, _, _ = special_args(
            self.fn, inputs=[message, history, *args], request=request
        )

        current_response = {"role": "assistant", "content": ""}
        if self.is_async:
            generator = self.fn(*inputs)
        else:
            generator = await anyio.to_thread.run_sync(
                self.fn, *inputs, limiter=self.limiter
            )
            generator = SyncToAsyncIterator(generator, self.limiter)
        try:
            first_response = self.response_as_dict(
                await async_iteration(generator), current_response=current_response
            )
            if self.multimodal and isinstance(message, MultimodalData):
                for x in message.files:
                    history.append(
                        {"role": "user", "content": cast(FileDataDict, x.model_dump())}
                    )
                update = history + [
                    {"role": "user", "content": message.text},
                    first_response,
                ]
                yield update, update
            else:
                update = history + [
                    {"role": "user", "content": message},
                    first_response,
                ]
                yield update, update
        except StopIteration:
            if self.multimodal and isinstance(message, MultimodalData):
                self._append_multimodal_history(message, None, history)
                yield history, history
            else:
                update = history + [{"role": "user", "content": message}]
                yield update, update
        async for response in generator:
            new_response = self.response_as_dict(
                response, current_response=current_response
            )
            print("NEW RESPONSE", new_response)
            if self.multimodal and isinstance(message, MultimodalData):
                update = history + [
                    {"role": "user", "content": message.text},
                    new_response,
                ]
                yield update, update
            else:
                update = history + [{"role": "user", "content": message}, new_response]
                yield update, update

    async def _api_submit_fn(
        self, message: str, history: list[MessageDict], request: Request, *args
    ) -> tuple[MessageDict, list[MessageDict]]:
        inputs, _, _ = special_args(
            self.fn, inputs=[message, history, *args], request=request
        )

        if self.is_async:
            response = await self.fn(*inputs)
        else:
            response = await anyio.to_thread.run_sync(
                self.fn, *inputs, limiter=self.limiter
            )
        new_response = self.response_as_dict(response)
        history.extend([{"role": "user", "content": message}, new_response])
        return new_response, history

    async def _api_stream_fn(
        self, message: str, history: list[list[str | None]], request: Request, *args
    ) -> AsyncGenerator:
        inputs, _, _ = special_args(
            self.fn, inputs=[message, history, *args], request=request
        )

        if self.is_async:
            generator = self.fn(*inputs)
        else:
            generator = await anyio.to_thread.run_sync(
                self.fn, *inputs, limiter=self.limiter
            )
            generator = SyncToAsyncIterator(generator, self.limiter)
        try:
            first_response = self.response_as_dict(await async_iteration(generator))
            yield (
                first_response,
                history + [{"role": "user", "content": message}, first_response],
            )
        except StopIteration:
            yield None, history + [[message, None]]
        async for response in generator:
            new_response = self.response_as_dict(response)
            yield (
                response,
                history
                + history
                + [{"role": "user", "content": message}, new_response],
            )

    async def _examples_fn(self, message: str, *args) -> list[list[str | None]]:
        inputs, _, _ = special_args(self.fn, inputs=[message, [], *args], request=None)

        if self.is_async:
            response = await self.fn(*inputs)
        else:
            response = await anyio.to_thread.run_sync(
                self.fn, *inputs, limiter=self.limiter
            )
        return [[message, response]]

    async def _examples_stream_fn(
        self,
        message: str,
        *args,
    ) -> AsyncGenerator:
        inputs, _, _ = special_args(self.fn, inputs=[message, [], *args], request=None)

        if self.is_async:
            generator = self.fn(*inputs)
        else:
            generator = await anyio.to_thread.run_sync(
                self.fn, *inputs, limiter=self.limiter
            )
            generator = SyncToAsyncIterator(generator, self.limiter)
        async for response in generator:
            yield [[message, response]]

    async def _delete_prev_fn(
        self,
        message: str | MultimodalData,
        history: list[MessageDict],
    ) -> tuple[
        list[MessageDict],
        str | MultimodalData,
        list[MessageDict],
    ]:
        if self.multimodal and isinstance(message, MultimodalData):
            remove_input = (
                len(message.files) + 1
                if message.text is not None
                else len(message.files)
            )
            history = history[:-remove_input]
        else:
            history = history[:-1]
        return history, message or "", history



import gradio as gr
from fastapi import FastAPI, Request
from fastapi.templating import Jinja2Templates
import uvicorn
from pathlib import Path
from dotenv import load_dotenv

load_dotenv()


from hf_agent_demo import demo as hf_agent
from hf_chatinterface import demo as hf_chatinterface
from langchain_agent_demo import demo as langchain_agent
from openai_stream_demo import demo as openai_stream
from open_ai_function_calling_demo import demo as openai_agent


def all_demos():
    for file_name, demo, name in [
        ("hf_chatinterface", hf_chatinterface, "HF Chat Interface 🤗"),
        ("openai_stream_demo", openai_stream, "OpenAI API Chatbot"),
        ("hf_agent_demo", hf_agent, "Transformers Agent 🤖"),
        ("langchain_agent_demo", langchain_agent, "LangChain Agent 🦜⛓️"),
        ("open_ai_function_calling_demo", openai_agent, "OpenAI Function Calling"),
    ]:
        with gr.Blocks() as new_demo:
            with gr.Tabs():
                with gr.Tab("Demo"):
                    demo.render()
                with gr.Tab("Code"):
                    gr.Code(
                        value=Path(f"{file_name}.py").read_text(), language="python"
                    )
        yield new_demo, file_name, name


demo_and_name = [tup for tup in all_demos()]


app = FastAPI()

templates = Jinja2Templates(directory="templates")


@app.get("/")
def index(request: Request):
    names = [(p[2], p[1]) for p in demo_and_name]
    return templates.TemplateResponse(
        "index.html",
        {
            "request": request,
            "names": names,
            "initial_demo": names[0][1],
            "is_space": False,
        },
    )


for demo, name, _ in demo_and_name:
    app = gr.mount_gradio_app(app, demo, f"/demo/{name}")


if __name__ == "__main__":
    uvicorn.run(app, port=7860, host="0.0.0.0")



"""gr.Chatbot() component."""

from __future__ import annotations

from pathlib import Path
from typing import Any, Callable, Literal, cast

from gradio_client import utils as client_utils

from gradio import utils
from gradio.components.base import Component
from gradio.events import Events
from gradio.processing_utils import move_resource_to_block_cache
from .utils import (
    Message,
    ChatbotData,
    FileData,
    FileMessage,
    MessageDict,
    ApiReturnObj,
)


class AgentChatbot(Component):
    """
    Creates a chatbot that displays user-submitted messages and responses. Supports a subset of Markdown including bold, italics, code, tables.
    Also supports audio/video/image files, which are displayed in the AgentChatbot, and other kinds of files which are displayed as links. This
    component is usually used as an output component.

    Demos: chatbot_simple, chatbot_multimodal
    Guides: creating-a-chatbot
    """

    EVENTS = [Events.change]
    data_model = ChatbotData

    def __init__(
        self,
        value: list[dict[str, Any]] | Callable | None = None,
        *,
        label: str | None = None,
        every: float | None = None,
        show_label: bool | None = None,
        container: bool = True,
        scale: int | None = None,
        min_width: int = 160,
        visible: bool = True,
        elem_id: str | None = None,
        elem_classes: list[str] | str | None = None,
        render: bool = True,
        key: int | str | None = None,
        height: int | str | None = None,
        latex_delimiters: list[dict[str, str | bool]] | None = None,
        rtl: bool = False,
        show_share_button: bool | None = None,
        show_copy_button: bool = False,
        avatar_images: tuple[str | Path | None, str | Path | None] | None = None,
        sanitize_html: bool = True,
        render_markdown: bool = True,
        bubble_full_width: bool = True,
        line_breaks: bool = True,
        likeable: bool = False,
        layout: Literal["panel", "bubble"] | None = None,
        placeholder: str | None = None,
    ):
        """
        Parameters:
            value: Default value to show in chatbot. If callable, the function will be called whenever the app loads to set the initial value of the component.
            label: The label for this component. Appears above the component and is also used as the header if there are a table of examples for this component. If None and used in a `gr.Interface`, the label will be the name of the parameter this component is assigned to.
            every: If `value` is a callable, run the function 'every' number of seconds while the client connection is open. Has no effect otherwise. The event can be accessed (e.g. to cancel it) via this component's .load_event attribute.
            show_label: if True, will display label.
            container: If True, will place the component in a container - providing some extra padding around the border.
            scale: relative size compared to adjacent Components. For example if Components A and B are in a Row, and A has scale=2, and B has scale=1, A will be twice as wide as B. Should be an integer. scale applies in Rows, and to top-level Components in Blocks where fill_height=True.
            min_width: minimum pixel width, will wrap if not sufficient screen space to satisfy this value. If a certain scale value results in this Component being narrower than min_width, the min_width parameter will be respected first.
            visible: If False, component will be hidden.
            elem_id: An optional string that is assigned as the id of this component in the HTML DOM. Can be used for targeting CSS styles.
            elem_classes: An optional list of strings that are assigned as the classes of this component in the HTML DOM. Can be used for targeting CSS styles.
            render: If False, component will not render be rendered in the Blocks context. Should be used if the intention is to assign event listeners now but render the component later.
            key: if assigned, will be used to assume identity across a re-render. Components that have the same key across a re-render will have their value preserved.
            height: The height of the component, specified in pixels if a number is passed, or in CSS units if a string is passed.
            latex_delimiters: A list of dicts of the form {"left": open delimiter (str), "right": close delimiter (str), "display": whether to display in newline (bool)} that will be used to render LaTeX expressions. If not provided, `latex_delimiters` is set to `[{ "left": "$$", "right": "$$", "display": True }]`, so only expressions enclosed in $$ delimiters will be rendered as LaTeX, and in a new line. Pass in an empty list to disable LaTeX rendering. For more information, see the [KaTeX documentation](https://katex.org/docs/autorender.html).
            rtl: If True, sets the direction of the rendered text to right-to-left. Default is False, which renders text left-to-right.
            show_share_button: If True, will show a share icon in the corner of the component that allows user to share outputs to Hugging Face Spaces Discussions. If False, icon does not appear. If set to None (default behavior), then the icon appears if this Gradio app is launched on Spaces, but not otherwise.
            show_copy_button: If True, will show a copy button for each chatbot message.
            avatar_images: Tuple of two avatar image paths or URLs for user and bot (in that order). Pass None for either the user or bot image to skip. Must be within the working directory of the Gradio app or an external URL.
            sanitize_html: If False, will disable HTML sanitization for chatbot messages. This is not recommended, as it can lead to security vulnerabilities.
            render_markdown: If False, will disable Markdown rendering for chatbot messages.
            bubble_full_width: If False, the chat bubble will fit to the content of the message. If True (default), the chat bubble will be the full width of the component.
            line_breaks: If True (default), will enable Github-flavored Markdown line breaks in chatbot messages. If False, single new lines will be ignored. Only applies if `render_markdown` is True.
            likeable: Whether the chat messages display a like or dislike button. Set automatically by the .like method but has to be present in the signature for it to show up in the config.
            layout: If "panel", will display the chatbot in a llm style layout. If "bubble", will display the chatbot with message bubbles, with the user and bot messages on alterating sides. Will default to "bubble".
            placeholder: a placeholder message to display in the chatbot when it is empty. Centered vertically and horizontally in the AgentChatbot. Supports Markdown and HTML. If None, no placeholder is displayed.
        """
        self.likeable = likeable
        self.height = height
        self.rtl = rtl
        if latex_delimiters is None:
            latex_delimiters = [{"left": "$$", "right": "$$", "display": True}]
        self.latex_delimiters = latex_delimiters
        self.show_share_button = (
            (utils.get_space() is not None)
            if show_share_button is None
            else show_share_button
        )
        self.render_markdown = render_markdown
        self.show_copy_button = show_copy_button
        self.sanitize_html = sanitize_html
        self.bubble_full_width = bubble_full_width
        self.line_breaks = line_breaks
        self.layout = layout
        super().__init__(
            label=label,
            every=every,
            show_label=show_label,
            container=container,
            scale=scale,
            min_width=min_width,
            visible=visible,
            elem_id=elem_id,
            elem_classes=elem_classes,
            render=render,
            key=key,
            value=value,
        )
        self.avatar_images: list[dict | None] = [None, None]
        if avatar_images is None:
            pass
        else:
            self.avatar_images = [
                self.serve_static_file(avatar_images[0]),
                self.serve_static_file(avatar_images[1]),
            ]
        self.placeholder = placeholder

    def preprocess(
        self,
        payload: ChatbotData | None,
    ) -> list[MessageDict]:
        """
        Parameters:
            payload: data as a ChatbotData object
        Returns:
            Passes the messages in the chatbot as a `list[list[str | None | tuple]]`, i.e. a list of lists. The inner list has 2 elements: the user message and the response message. Each message can be (1) a string in valid Markdown, (2) a tuple if there are displayed files: (a filepath or URL to a file, [optional string alt text]), or (3) None, if there is no message displayed.
        """
        if payload is None:
            return []
        return payload.model_dump()

    def _postprocess_chat_messages(
        self, chat_message: Message | MessageDict | ApiReturnObj
    ) -> list[Message]:
        if not isinstance(chat_message, (Message, dict)) and Message.matches_protocol(
            chat_message
        ):
            chat_message = Message.from_api(chat_message)

        if isinstance(chat_message, dict):
            chat_message = Message(**chat_message)  # pyright: ignore
        if isinstance(chat_message.content, FileMessage):
            cached_file = move_resource_to_block_cache(
                chat_message.content.file.path, block=self
            )
            assert cached_file is not None
            chat_message.content.file.path = cached_file
            return [cast(Message, chat_message)]

        # extract file path from message
        new_messages = []
        for word in chat_message.content.split(" "):
            if (filepath := Path(word)).exists() and filepath.is_file():
                filepath = move_resource_to_block_cache(filepath, block=self)
                assert filepath is not None
                mime_type = client_utils.get_mimetype(filepath)
                new_messages.append(
                    Message(
                        role=chat_message.role,
                        metadata=cast(Message, chat_message).metadata,
                        content=FileMessage(
                            file=FileData(path=filepath, mime_type=mime_type)
                        ),
                    )
                )

        return [cast(Message, chat_message), *new_messages]

    def postprocess(
        self,
        value: list[Message | MessageDict | ApiReturnObj] | None,
    ) -> ChatbotData:
        """
        Parameters:
            value: expects a `list[list[str | None | tuple]]`, i.e. a list of lists. The inner list should have 2 elements: the user message and the response message. The individual messages can be (1) strings in valid Markdown, (2) tuples if sending files: (a filepath or URL to a file, [optional string alt text]) -- if the file is image/video/audio, it is displayed in the AgentChatbot, or (3) None, in which case the message is not displayed.
        Returns:
            an object of type ChatbotData
        """
        if value is None:
            return ChatbotData(root=[])
        processed_messages = [
            msg for message in value for msg in self._postprocess_chat_messages(message)
        ]
        return ChatbotData(root=processed_messages)

    def example_payload(self) -> Any:
        return [
            Message(role="user", content="Hello!").model_dump(),
            Message(role="assistant", content="How can I help you?").model_dump(),
        ]

    def example_value(self) -> Any:
        return [
            Message(role="user", content="Hello!"),
            Message(role="assistant", content="How can I help you?"),
        ]
